"""The `tokamak` module is part of the `dress` package and contains 
tools for facilitating reaction product spectra calculations in tokamak
geometry, i.e. a toroidal system with toroidal symmetry, where quanteties
are typically a function either of (R,Z) or of a flux surface label ρ."""

import dress.tokamak.tokadist
import dress.tokamak.tokapart
import dress.tokamak.utils

